import numpy as np


def isnan(x) -> bool:
    """checks if dataframe value is nan or None.
    Takes into account different possibilities (e.g np.nan, float('NaN'))"""
    return x is np.nan or x != x or x is None



import json
from typing import List
import numpy as np
import pandas as pd
from strsimpy.weighted_levenshtein import WeightedLevenshtein
from symspellpy import SymSpell, Verbosity

from mim_ocr import mim_ocr_cfg, check_mim_ocr_cfg_exists
from mim_ocr.heuristics import Feature, Occurrence
from mim_ocr.utils import isnan


POLISH_LETTER_COST = 0.1
POLISH_LETTERS_LINKS = {
    'Ą': 'A',
    'ą': 'a',
    'Ę': 'E',
    'ę': 'e',
    'Ł': 'L',
    'ł': 'l',
    'Ć': 'C',
    'ć': 'c',
    'Ń': 'N',
    'ń': 'n',
    'Ó': 'O',
    'ó': 'o',
    'Ś': 'S',
    'ś': 's',
    'Ź': 'Z',
    'ź': 'z',
    'Ż': 'Z',
    'ż': 'z',
}
# Maximum difference between keyword length and matched test lenghts (use for speedup)
# Higher value will allow more fuzzy results: e.g. with more additional dots
ALLOWED_WORD_LEN_DIFFERENCE: int = 2

# Length of the string at which distance 1.0 should be accepted
MAXIMUM_DISTANCE_LENGTH_WITH_COEFFICIENT_1 = 4.0

# Power which used to scale the allowed distance.
# MAXIMUM_DISTANCE_POWER = 0.5 means that if you make string 4 times longer
# the allowed distance will double
MAXIMUM_DISTANCE_POWER = 0.5

# Maximum ord() of precalculated weighted distance matrix.
LEVEN_PARAMS_SIZE = 128

# Maximum number of words allowed in text
MAX_WORD_NUMBER = 20

SYMSPELL_PREFIX_LENGTH = 7 # noqa see https://towardsdatascience.com/symspell-vs-bk-tree-100x-faster-fuzzy-string-search-spell-checking-c4f10d80a078 fro explanation
SYMSPELL_MAX_DISTANCE = 3
SYMSEPELL_MAX_LEN_DISTANCE_1 = 3
SYMSEPELL_MAX_LEN_DISTANCE_2 = 6


check_mim_ocr_cfg_exists()
with open(mim_ocr_cfg.features.fuzzy_matching.params_weighted_leven_path, "r") as f:
    leven_params = json.load(f)
    for k in leven_params.keys():
        if k == "comment":
            continue
        assert len(leven_params[k]) == LEVEN_PARAMS_SIZE
        leven_params[k] = np.array(leven_params[k])


def get_single_operation_cost(key: str, char: str):
    if key not in ('insert_costs', 'delete_costs'):
        raise ValueError("Incorrect key.")
    if len(char) > 1:
        raise ValueError("Only single char accepted.")
    o = ord(char)
    if o < LEVEN_PARAMS_SIZE:
        return leven_params[key][o]
    if char in POLISH_LETTERS_LINKS:
        return leven_params[key][ord(POLISH_LETTERS_LINKS[char])]+POLISH_LETTER_COST
    return 1.0


def get_insertion_cost(char: str):
    return get_single_operation_cost('insert_costs', char)


def get_deletion_cost(char: str):
    return get_single_operation_cost('delete_costs', char)


def get_substitution_cost(char_a: str, char_b: str):
    o_a = ord(char_a)
    o_b = ord(char_b)
    if o_a < LEVEN_PARAMS_SIZE and o_b < LEVEN_PARAMS_SIZE:
        return leven_params['substitute_costs'][o_a][o_b]
    if char_a in POLISH_LETTERS_LINKS:
        if char_b == POLISH_LETTERS_LINKS[char_a]:
            return POLISH_LETTER_COST
        return get_substitution_cost(POLISH_LETTERS_LINKS[char_a], char_b) + POLISH_LETTER_COST
    if char_b in POLISH_LETTERS_LINKS:
        if char_a == POLISH_LETTERS_LINKS[char_b]:
            return POLISH_LETTER_COST
        return get_substitution_cost(char_a, POLISH_LETTERS_LINKS[char_b]) + POLISH_LETTER_COST
    return 1.0


pl_weighted_levenshtein = WeightedLevenshtein(
    substitution_cost_fn=get_substitution_cost,
    insertion_cost_fn=get_insertion_cost,
    deletion_cost_fn=get_deletion_cost)


class FuzzyMatchKeywordFeature(Feature):
    """
    Feature that looks for keywords using weighted-levenshtein algorithm.
    Weight are taken from https://github.com/zas97/ocr_weighted_levenshtein

    Good description on how does weighted levenshtein algorithm works can be found on
    https://github.com/infoscout/weighted-levenshtein
    """
    def __init__(self, name: str, keywords: List[str], priority: int = 5,
                 allow_upper: bool = False):

        super().__init__(name, priority=priority)

        if allow_upper:
            keywords = list(set(keywords+[keyword.upper() for keyword in keywords]))
        keywords.sort()

        self.sym_spell = SymSpell(max_dictionary_edit_distance=SYMSPELL_MAX_DISTANCE,
                                  prefix_length=SYMSPELL_PREFIX_LENGTH)

        for keyword in keywords:
            self.sym_spell.create_dictionary_entry(keyword, 1)

        self.keywords_by_words_count_and_length: dict[int, dict[int, [List[str]]]] = \
            {i: {} for i in range(MAX_WORD_NUMBER)}

        for keyword in keywords:
            nr_words = len(keyword.split(" "))
            keyword_len = len(keyword)
            self.keywords_by_words_count_and_length[nr_words].setdefault(keyword_len, []).append(keyword)

    def find_occurrences(self, text: str) -> List[Occurrence]:
        if "  " in text:
            raise ValueError("Multiple spaces are not supported.")

        occurrences: List[Occurrence] = []

        words = text.split(" ")
        nr_words = len(words)
        for word_frame_length in range(1, MAX_WORD_NUMBER+1):
            current_start_position = 0
            for i in range(nr_words-word_frame_length+1):
                t = " ".join(words[i:i + word_frame_length])
                occurrences += self.\
                    find_occurrences_for_word_frame_and_position(search_text=t,
                                                                 word_frame_length=word_frame_length,
                                                                 current_start_position=current_start_position,
                                                                 full_text=text)
                current_start_position += len(words[i])+1
        return occurrences

    def find_occurrences_for_word_frame_and_position(self, search_text: str, word_frame_length: int,
                                                     current_start_position: int,
                                                     full_text: str) -> List[Occurrence]:
        """Distance between search text and keywords is calculated using weighted Levenstein algorithm.
        Maximum allowed distance is proportional to MAXIMUM_DISTANCE_POWER of the search word length.

        Fuzzy mathing is not allowed for strings of length 1.
        """

        occurrences: List[Occurrence] = []

        if len(search_text) <= SYMSEPELL_MAX_LEN_DISTANCE_1:
            max_symspell_distance = 1
        elif len(search_text) <= SYMSEPELL_MAX_LEN_DISTANCE_2:
            max_symspell_distance = 2
        else:
            max_symspell_distance = 3

        if not self.sym_spell.lookup(search_text, Verbosity.CLOSEST, max_edit_distance=max_symspell_distance):
            return occurrences

        t_length = len(search_text)
        current_end_position = current_start_position + t_length - 1
        for j in range(max(1, t_length - ALLOWED_WORD_LEN_DIFFERENCE), t_length + ALLOWED_WORD_LEN_DIFFERENCE):
            for keyword in self.keywords_by_words_count_and_length.get(word_frame_length, {}).get(j, []):
                distance = pl_weighted_levenshtein.distance(search_text, keyword)
                if distance == 0 or (((distance * MAXIMUM_DISTANCE_LENGTH_WITH_COEFFICIENT_1 / t_length)
                                      ** MAXIMUM_DISTANCE_POWER <= 1) and t_length > 1):
                    occurrences.append(Occurrence(full_text=full_text, feature_name=self.name,
                                                  start=current_start_position, end=current_end_position,
                                                  matched_text=search_text, priority=self.priority,
                                                  additional_info=keyword))
        return occurrences


class CSVFuzzyMatchKeywordFeature(FuzzyMatchKeywordFeature):
    """Any CSV format that can be read by Pandas is supported.
    every cell in the first sheet (except for the header) are treated as keywords."""
    def __init__(self, name: str, csv_path: str, priority: int = 5,
                 allow_upper: bool = False):
        keywords_df = pd.read_csv(csv_path)
        keywords = keywords_df.values.tolist()
        keyword_list = [item.strip() for sublist in keywords for item in sublist if not isnan(item)]

        super().__init__(name, priority=priority, keywords=keyword_list, allow_upper=allow_upper)



import json
import re
import uuid
import jsonpickle
from enum import Enum
from pathlib import Path
from os import PathLike
from typing import Dict, List, Tuple, Optional, Iterator, Any, Union

import cv2
import numpy as np
import pandas as pd
from pptree import pptree


class BoxType(Enum):
    ROOT_BOX = 0
    TESSERACT_DOCUMENT = 1
    TESSERACT_PAGE = 2
    TESSERACT_PARAGRAPH = 3
    TESSERACT_LINE = 4
    TESSERACT_WORD = 5
    GCP_DOCUMENT = 17
    # GCP_BLOCK_PICTURE = 18  # TODO:
    # GCP_BLOCK_RULER = 19  # TODO:
    GCP_BLOCK_TABLE = 20
    GCP_BLOCK_TEXT = 21
    # GCP_BLOCK_UNKNOWN = 22  # TODO:
    # GCP_BLOCK_BARCODE = 23  # TODO:
    GCP_BLOCK_PARAGRAPH = 24
    GCP_BLOCK_WORD = 25
    CUSTOM = 30
    PREDICTED_PAGE = 90
    AWS_BLOCK_PAGE = 100
    AWS_BLOCK_LINE = 101
    AWS_BLOCK_WORD = 102


PRECEDING_BOX_TYPES: Dict[BoxType, Optional[BoxType]] = {
    BoxType.ROOT_BOX: None,
    BoxType.TESSERACT_DOCUMENT: BoxType.ROOT_BOX,
    BoxType.TESSERACT_PAGE: BoxType.TESSERACT_DOCUMENT,
    BoxType.TESSERACT_PARAGRAPH: BoxType.TESSERACT_PAGE,
    BoxType.TESSERACT_LINE: BoxType.TESSERACT_PARAGRAPH,
    BoxType.TESSERACT_WORD: BoxType.TESSERACT_LINE,
    BoxType.PREDICTED_PAGE: BoxType.ROOT_BOX,
    BoxType.GCP_DOCUMENT: BoxType.ROOT_BOX,
    BoxType.GCP_BLOCK_TEXT: BoxType.GCP_DOCUMENT,
    BoxType.GCP_BLOCK_PARAGRAPH: BoxType.GCP_BLOCK_TEXT,
    BoxType.GCP_BLOCK_WORD: BoxType.GCP_BLOCK_PARAGRAPH,
}


class Box:
    def __init__(self,
                 text: Optional[str],
                 box_type: BoxType,
                 conf: Optional[float] = None,
                 left: int = 0, top: int = 0, right: int = 0, bottom: int = 0,
                 parent: Optional['Box'] = None,
                 additional_data: Optional[Dict[str, Any]] = None,
                 box_id: Optional[str] = None,
                 ) -> None:
        """
        The main class for storing OCR results.

        The objects represents boxes with text. Thanks to 'parent' parameter the boxes can be
        structured into tree.

        Args:
            conf (Optional[float]): A confidence of OCR (between 0 and 100)
            additional_data (Optional[Dict[str, Any]]): dictionary with extra data associated with node.
            box_id: Optional unique identifier. Useful to identify e.x.
                    words after saving box to disk in batch processing
        """
        if box_type == BoxType.ROOT_BOX:
            self.box_id = None
        elif box_id is None:
            self.box_id = str(uuid.uuid4())
        else:
            self.box_id = str(box_id)
        self.left = left
        self.top = top
        self.right = right
        self.bottom = bottom
        self.conf = conf
        self.text = text or ''
        self.box_type = box_type
        self.children: List['Box'] = []
        self.parent = parent
        self.additional_data = additional_data if additional_data is not None else {}

        # dictionary allowing to quickly access any box using its id will be filled only for ROOT_BOX
        self.box_dict: Optional[Dict[int, 'Box']] = None
        if box_type == BoxType.ROOT_BOX:
            self.box_dict = {}

    def __repr__(self) -> str:
        return (f"Rect(left={self.left}, top={self.top}, right={self.right}, bottom={self.bottom}), "
                f"Text: '{self.text}', BoxType: {self.box_type}, n_children={len(self.children)}")

    __str__ = __repr__

    def __eq__(self, other):
        if self.to_dict() != other.to_dict():
            return False
        return self.children == other.children

    def height(self) -> int:
        return self.bottom - self.top

    def width(self) -> int:
        return self.right - self.left

    def size(self) -> Tuple[int, int]:
        return self.width(), self.height()

    def print(self):
        return pptree.pptree(self)

    def draw_border(self, img: np.ndarray, color: Union[Tuple[int, int, int], List[int]]) -> np.ndarray:
        return cv2.rectangle(img, (self.left, self.top), (self.right, self.bottom), color=color)

    def preorder_traversal(self) -> Iterator['Box']:

        yield self
        for child in self.children:
            for box in child.preorder_traversal():
                yield box

    def reverse_order_traversal(self) -> Iterator['Box']:
        for child in self.children[::-1]:
            for box in child.reverse_order_traversal():
                yield box
        yield self

    def add_box_based_on_type(self, added_box: 'Box') -> None:
        """Add a new box to tree with a parent based on box type"""

        added_type = added_box.box_type
        preceding_type = PRECEDING_BOX_TYPES[added_type]

        for box in self.reverse_order_traversal():
            box_type = box.box_type
            if box_type == preceding_type:
                Box.add_child(box, added_box)
                return
        raise ValueError("Unable to insert box. No suitable parent box found.")

    def calc_confidence(self) -> Dict[str, float]:
        """Returns dict with confidence statistics of words in boxes:
           avg_confidence, total_letters, avg_letters_per_box."""

        df = self.to_dataframe()
        if df.empty:
            return {
                'avg_confidence': 0.0,
                'total_letters': 0.0,
                'avg_letters_per_box': 0.0,
            }
        df = df[df['has_children'] == 0]
        total_letters = df['text'].str.len().sum()
        total_boxes = len(df)
        avg_confidence_weighted_by_text_lengths = (df['conf'] * df['text'].str.len()).sum() / (total_letters or 1)
        return {
            'avg_confidence': avg_confidence_weighted_by_text_lengths,
            'total_letters': total_letters,
            'avg_letters_per_box': total_letters / (total_boxes or 1)
        }

    def to_dict(self) -> Dict:
        d = {
            'left': self.left,
            'top': self.top,
            'right': self.right,
            'bottom': self.bottom,
            'conf': self.conf,
            'text': self.text or '',
            'has_children': len(self.children) > 0,
            'box_type': self.box_type.value,
            'box_id': self.box_id,
        }
        d.update(self.additional_data)
        return d

    def to_json_file(self, output_path: Union[PathLike, str]) -> None:
        json_string = jsonpickle.encode(self)
        # before exporting, we convert json_string to json_object, so the output file can be easily read by human
        json_object = json.loads(json_string)
        with open(output_path, 'w') as output_file:
            json.dump(json_object, output_file, indent=2)

    @staticmethod
    def from_json_file(input_path: Union[PathLike, str]) -> 'Box':
        with open(input_path) as input_file:
            input_data = json.load(input_file)
        input_json_string = json.dumps(input_data)
        return Box.from_json_str(input_json_string)

    @staticmethod
    def from_json_str(input_json_string: str):
        box = jsonpickle.decode(input_json_string)
        box._recalculate_box_dict()
        return box

    def _recalculate_box_dict(self):
        if self.box_type != BoxType.ROOT_BOX:
            raise ValueError("You can recalculate box_dict only for Root Box.")
        self.box_dict = {b.box_id: b for b in self.preorder_traversal()}

    def to_dataframe(self) -> pd.DataFrame:
        if self.box_type == BoxType.ROOT_BOX and not self.children:
            return pd.DataFrame(columns=["left", "right", "bottom", "top", "conf", "text", "has_children", "box_type"])
        df = pd.DataFrame.from_records(self.to_list())
        df.set_index("box_id", inplace=True)
        return df

    def to_list(self) -> List[Dict]:
        return [b.to_dict() for b in self.preorder_traversal() if (b.box_type != BoxType.ROOT_BOX)]

    def to_csv(self, path: str) -> None:
        self.to_dataframe().to_csv(path)

    def to_excel(self, path: str) -> None:
        self.to_dataframe().to_excel(path)

    @staticmethod
    def create_root_box() -> 'Box':
        """Creates dummy box with no dimensions, text, confidence etc. that will be a root for real boxes"""
        return Box(left=-1, top=-1, right=-1, bottom=-1, conf=-1,
                   text=None, box_type=BoxType.ROOT_BOX, parent=None)

    @staticmethod
    def create_page_box(page_number: int = 0, page_size: Tuple[int, int] = None) -> 'Box':
        return Box(left=-1, top=-1, right=-1, bottom=-1, conf=-1,
                   text=None, box_type=BoxType.PREDICTED_PAGE, parent=None,
                   additional_data={"page_number": page_number, "page_size": page_size})

    def get_subboxes(self, box_type: Optional[BoxType] = None) -> List['Box']:
        """
        returns list of subboxes with given type
        if type is not specified all subboxes are returned
        """
        box_iterator = self.preorder_traversal()
        next(box_iterator)
        if box_type:
            return [box for box in box_iterator if box.box_type == box_type]
        return [box for box in box_iterator]

    @staticmethod
    def add_child(parent: 'Box', child: 'Box') -> None:
        parent.children.append(child)
        child.parent = parent
        root_box = parent.get_root()
        for box in child.preorder_traversal():
            if box.box_type == BoxType.ROOT_BOX:
                raise ValueError("You cannot have two root boxes in one tree")
            root_box.box_dict[box.box_id] = box

    @staticmethod
    def from_dataframe(df: pd.DataFrame) -> 'Box':
        """
        create Box from pandas DataFrame
        required index: it will be used as box_id
        required columns (at the beginning):
            left: int
            top: int
            right: int
            bottom: int
            conf: float
            box_type: int - compatible with BoxType enum
        optional columns:
            : values of additional data. might be an integer, float or a string.
        """
        fixed_columns = ['left', 'top', 'right', 'bottom', 'conf', 'text', 'has_children', 'box_type']
        assert set(list(df.columns)[:len(fixed_columns)]) == set(fixed_columns)

        for colname in fixed_columns:
            assert colname in df.columns

        additional_columns_names = list(df.columns)[len(fixed_columns):]

        tree = Box.create_root_box()
        for index, row in df.iterrows():
            additional_data = {column_name: row[column_name] for column_name in additional_columns_names}
            new_box = Box(left=row.left, top=row.top, right=row.right, bottom=row.bottom,
                          conf=row.conf,
                          text=row.text,
                          box_type=BoxType(row.box_type),
                          box_id=index,
                          additional_data=additional_data)
            tree.add_box_based_on_type(new_box)

        return tree

    @staticmethod
    def from_excel(path: Path) -> 'Box':
        df = pd.read_excel(path, keep_default_na=False, index_col=0, converters={0: str, "text": str})
        return Box.from_dataframe(df)

    @staticmethod
    def from_csv(path: Path) -> 'Box':
        df = pd.read_csv(
            path, keep_default_na=False, index_col=0,
            converters={0: str, "text": str}, on_bad_lines="warn")
        return Box.from_dataframe(df)

    def add_pages(self, boxes: List['Box']):
        '''Append PREDICTED_PAGE boxes present as children of input boxes.'''
        for box in boxes:
            for page_box in box.children:
                if page_box.box_type != BoxType.PREDICTED_PAGE:
                    raise ValueError("Wrong box format.")
                page_box.additional_data['page_number'] = len(self.children)
                self.add_child(self, page_box)

    def get_full_text(self):
        return " ".join([b.text for b in self.preorder_traversal() if b.text])

    def get_root(self) -> 'Box':
        if not self.parent:
            return self
        return self.parent.get_root()

    def get_ancestor_box_additional_data(self, additional_data_key: str) -> Tuple[int, int]:
        """
        If the box has additional_data_key in it's additional_data, returns the value attached to it.
        Else, returns such value of it's closest ancestor or None.
        """
        if additional_data_key in self.additional_data:
            return self.additional_data[additional_data_key]
        if self.parent is not None:
            return self.parent.get_ancestor_box_additional_data(additional_data_key)
        return None

    def full_box_height(self) -> int:
        min_full_bot_top = min([b.top for b in self.preorder_traversal() if b.box_type != BoxType.ROOT_BOX])
        max_full_box_bottom = max([b.bottom for b in self.preorder_traversal() if b.box_type != BoxType.ROOT_BOX])
        return max_full_box_bottom - min_full_bot_top

    def full_box_width(self) -> int:
        min_full_box_left = min([b.left for b in self.preorder_traversal() if b.box_type != BoxType.ROOT_BOX])
        max_full_box_right = max([b.right for b in self.preorder_traversal() if b.box_type != BoxType.ROOT_BOX])
        return max_full_box_right - min_full_box_left

    def has_any_text(self) -> bool:
        return bool(re.search(r"\S", self.get_full_text()))

    def box_number_in_parent(self) -> int:
        """Returns the position of box in tree among siblings"""
        if not self.parent:
            return 0
        position = 0
        while self.parent.children[position] is not self:
            position += 1

        return position

    @property
    def feature(self) -> Optional[str]:
        return self.additional_data.get("feature")

    @property
    def main_feature(self) -> Optional[str]:
        if not self.additional_data.get("feature"):
            return None
        return self.additional_data["feature"].rstrip("<->")

    def get_subbox_by_id(self, box_id: str) -> 'Box':
        for b in self.preorder_traversal():
            if b.box_id == box_id:
                return b
        raise ValueError("Box with given id not found")

    def merge_subboxes(self, i: int, j: int):
        """merge two child boxes"""
        child1 = self.children[i]
        child2 = self.children[j]

        child1.text = child1.text + child2.text
        child1.left = min(child1.left, child2.left)
        child1.right = max(child1.right, child2.right)
        child1.top = min(child1.top, child2.top)
        child1.bottom = max(child1.bottom, child2.bottom)

        del self.children[j]



from typing import Optional
import numpy as np

from mim_ocr.backends import OCRBackend
from mim_ocr.data_model import Box
from mim_ocr.data_model.box import BoxType


def run_ocr_on_box(backend: OCRBackend, img: np.ndarray,
                   box: Box, box_type: Optional[BoxType] = None,
                   *args, **kwargs) -> Box:
    """
        Run OCR algorithm on the whole box structure.

        For each of subboxes we run OCR algorithm.

        Args:
            backend (OCRBackend)
            img (np.ndarray): An original image.
            box (Box): Root box computed on img.
            box_type (Optional[BoxType]): Run OCR only on boxes with this type.
    """
    assert box.box_type == BoxType.ROOT_BOX

    # reverse order is necessary, because we modify the children
    for b in box.reverse_order_traversal():

        if (box_type is not None) and (b.box_type != box_type):
            continue

        if b.box_type == BoxType.ROOT_BOX:
            continue

        run_ocr_on_single_box(backend, img, b, *args, **kwargs)

    return box


# In OCR-37 move as a method of OCRBackend class.
def run_ocr_on_single_box(backend: OCRBackend, img: np.ndarray,
                          box: Box, *args, **kwargs) -> Box:
    """ Run OCR algorithm on single box.

        When the new structure is identified on the box, we replace
        the children of the box by the newly identified children and reset the old text.
    """
    new_b = backend.run_ocr_to_box(img[box.top:box.bottom, box.left:box.right, :], *args, **kwargs)

    # We have to update the identified coordinates
    for b_elem in new_b.preorder_traversal():
        b_elem.top = box.top + b_elem.top
        b_elem.bottom = box.top + b_elem.bottom
        b_elem.left = box.left + b_elem.left
        b_elem.right = box.left + b_elem.right

    if len(new_b.children) > 0:
        box.children = new_b.children
        box.text = ''
        info_dict = {"recomputed_ocr": backend.__class__.__name__}
        box.additional_data.update(info_dict)

    return box



import itertools
from typing import List, Optional

from mim_ocr.data_model import Box
from .feature import Occurrence, Feature
from ..data_model.box import BoxType


def heuristic_examine_box_lines(box: Box, features_to_check: Optional[List[Feature]] = None) -> None:
    """
    Find features for box elements and store information in additional Data.
    Found features are added to additional_data field of the analyzed box.
    Args:
        box: Box
        features_to_check: list of feature that will be checked
    """
    if features_to_check is None:
        features_to_check = []

    if not features_to_check:
        raise ValueError("Feature list required")

    for b in box.preorder_traversal():
        b.additional_data['feature'] = b.additional_data.get('feature', None)
        if b.box_type not in [BoxType.TESSERACT_LINE, BoxType.AWS_BLOCK_LINE, BoxType.GCP_BLOCK_PARAGRAPH]:
            continue
        text = b.get_full_text()
        occurrences = find_heuristic_features(text, features_to_check)
        if occurrences:
            _fill_box_with_feature_occurrences(b, occurrences)


def find_heuristic_features(original_text: str, features: Optional[List[Feature]] = None) -> List[Occurrence]:
    return list(itertools.chain.from_iterable([feature.find_occurrences(original_text) for feature in features]))


def _fill_box_with_feature_occurrences(tesseract_line_box: Box, feature_occurrences: List[Occurrence]) -> None:
    """As we calculated feature for text lines, we need to map them onto specific words e.g. for visualization"""
    for word_box in tesseract_line_box.children:
        if "feature" not in word_box.additional_data:
            word_box.additional_data["feature"] = None

    for feature_occurrence in sorted(feature_occurrences, key=lambda o: 256*o.priority + o.end - o.start, reverse=True):
        word_position_start = 0
        word_position_end = 0

        for word_box in tesseract_line_box.children:
            word_position_end += len(word_box.text)
            if word_box.feature \
                    and Feature.get_feature_priority_by_name(word_box.feature) >= feature_occurrence.priority:
                pass

            elif feature_occurrence.start >= word_position_start and feature_occurrence.end <= word_position_end:
                word_box.additional_data['feature'] = feature_occurrence.feature_name
            elif feature_occurrence.start < word_position_start <= word_position_end < feature_occurrence.end:
                word_box.additional_data['feature'] = feature_occurrence.feature_name + "-"
            elif word_position_start <= feature_occurrence.start <= word_position_end < feature_occurrence.end:
                word_box.additional_data['feature'] = feature_occurrence.feature_name + "<-"
            elif feature_occurrence.start < word_position_start <= feature_occurrence.end <= word_position_end:
                word_box.additional_data['feature'] = feature_occurrence.feature_name + "->"
            word_position_start = word_position_end = word_position_start + len(word_box.text) + 1



import math
from typing import Union, Tuple

import cv2
import numpy as np
from deskew import determine_skew
from scipy.ndimage import interpolation


def deskew_cv2(image: np.ndarray,
               method: str = "projection_profile",
               projection_delta: float = 0.5,
               projection_limit: int = 10,
               background: Union[int, Tuple[int, int, int]] = (0, 0, 0)) -> Tuple[np.ndarray, float]:
    if method == "hough_transform":
        best_angle = _determine_best_angle_hough_transform(image)
    elif method == "projection_profile":
        best_angle = _determine_best_angle_projection_profile(image, projection_delta, projection_limit)
    else:
        raise ValueError(f"Unknown deskew method: {method}")
    if best_angle == 0.0:
        return image, best_angle
    return rotate_cv2(image, best_angle, background), best_angle


def _determine_best_angle_hough_transform(image: np.ndarray) -> float:
    """See more: https://github.com/sbrunner/deskew"""
    if len(image.shape) == 3:
        img_grayscale = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    else:
        img_grayscale = image

    blur = cv2.medianBlur(img_grayscale, 3)
    thresh = cv2.threshold(blur, 0, 255, cv2.THRESH_BINARY_INV + cv2.THRESH_OTSU)[1]

    angle = determine_skew(thresh, num_angles=360, sigma=2.0,
                           num_peaks=40)  # small variation of default parameters obtained by trial and error
    if angle is None:
        return 0.0
    return float(angle)


def _determine_best_angle_projection_profile(image: np.ndarray, delta: float = .5, limit: int = 10) -> float:
    """Here's a modified implementation of the Projection Profile Method to correct skewed images
    as described in Projection profile based skew estimation algorithm for JBIG compressed images.
    After obtaining a binary image, the idea is to rotate the image at various angles and generate
    a histogram of pixels in each iteration. To determine the skew angle, we compare the maximum difference
    between peaks and using this skew angle, rotate the image to correct the skew. The amount of peaks
    to determine can be adjusted by the delta value, the lower the delta, the more peaks will be checked
    with the tradeoff that the process will take longer.
      - https://stackoverflow.com/questions/59660933/how-to-de-skew-a-text-image-also-retrieve-the-new-bounding-box-of-that-image  # noqa
      - http://www.cvc.uab.es/~bagdanov/pubs/ijdar98.pdf
    """

    gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    blur = cv2.medianBlur(gray, 3)
    thresh = cv2.threshold(blur, 0, 255, cv2.THRESH_BINARY_INV + cv2.THRESH_OTSU)[1]

    scores = []
    angles = np.arange(-limit, limit + delta, delta)
    for angle in angles:
        histogram, score = _determine_projection_profile_score(thresh, angle)
        scores.append(score)

    return angles[scores.index(max(scores))]


def _determine_projection_profile_score(arr, angle):
    data = interpolation.rotate(arr, angle, reshape=False, order=0)
    histogram = np.sum(data, axis=1)
    score = np.sum((histogram[1:] - histogram[:-1]) ** 2)
    return histogram, score


def get_rotation_matrix_cv2(image: np.ndarray, angle: float) -> Tuple[np.ndarray, float, float]:
    """Returs rotation_matrix and dimensions of result image (height, width)
    Source: deskew Readme (https://github.com/sbrunner/deskew)"""
    old_width, old_height = image.shape[:2]
    angle_radian = math.radians(angle)
    width = abs(np.sin(angle_radian) * old_height) + abs(np.cos(angle_radian) * old_width)
    height = abs(np.sin(angle_radian) * old_width) + abs(np.cos(angle_radian) * old_height)

    image_center = tuple(np.array(image.shape[1::-1]) / 2)
    rot_mat = cv2.getRotationMatrix2D(image_center, angle, 1.0)
    rot_mat[1, 2] += (width - old_width) / 2
    rot_mat[0, 2] += (height - old_height) / 2

    return rot_mat, height, width


def rotate_cv2(
        image: np.ndarray, angle: float, background: Union[int, Tuple[int, int, int]]
) -> np.ndarray:
    """Source: deskew Readme (https://github.com/sbrunner/deskew)"""

    # TODO: Check Another rotation method with different size, interpolation and border mode
    # (h, w) = image.shape[:2]
    # center = (w // 2, h // 2)
    # m = cv2.getRotationMatrix2D(center, best_angle, 1.0)
    # rotated = cv2.warpAffine(image, m, (w, h), flags=cv2.INTER_CUBIC, borderMode=cv2.BORDER_REPLICATE, )
    rot_mat, height, width = get_rotation_matrix_cv2(image, angle)

    return cv2.warpAffine(image, rot_mat, (int(round(height)), int(round(width))), borderValue=background)



from pathlib import Path
import loguru
import pandas as pd
from typing import List, Optional, Set, Any, Dict
import hyperscan
import re

from .feature import Feature, Occurrence
from mim_ocr.utils import isnan
from mim_ocr import mim_ocr_cfg, mim_ocr_cfg_required


class KeywordFeature(Feature):

    @mim_ocr_cfg_required
    def __init__(self, name: str, keywords: List[str], priority: int = 5,
                 allow_upper: bool = False, allow_first_letter_upper: bool = False,
                 allowed_letters_before_keyword_regex: str = "", allowed_letters_after_keyword_regex: str = "",
                 allow_recompute: bool = False):
        """
        Args:
            name:
            keywords: list of keywords that will be search in the text
            priority:
            allow_upper: if the feature is present in text if keyword is present in text in CAPITAL LETTERS.
            allow_first_letter_upper: if the feature is present in text if keyword is present with first Capital letter
            allowed_letters_before_keyword_regex: regex defining allowed text before keyword
            allowed_letters_after_keyword_regex: regex defining allowed text after keyword
            allow_recompute: Recompute hyperscan database every time. Used for automatic tests.
        """

        super().__init__(name=name, priority=priority)

        self.keywords = keywords
        self.allow_upper = allow_upper
        self.allow_first_letter_upper = allow_first_letter_upper
        self.allowed_letters_before_keyword_regex = allowed_letters_before_keyword_regex
        self.allowed_letters_after_keyword_regex = allowed_letters_after_keyword_regex
        hyperscan_db_path = Path(mim_ocr_cfg.features.keyword_features_hyperscan_databases_directory_path) / name
        if allow_recompute:
            self.compile(
                name=name, keywords=keywords,
                allow_upper=allow_upper,
                allow_first_letter_upper=allow_first_letter_upper,
                allowed_letters_before_keyword_regex=allowed_letters_before_keyword_regex,
                allowed_letters_after_keyword_regex=allowed_letters_after_keyword_regex
            )

        with open(hyperscan_db_path, 'rb') as hyperscan_db_file:
            self.hyperscan_database = hyperscan.loadb(hyperscan_db_file.read())

    @classmethod
    @mim_ocr_cfg_required
    def compile(
            cls, name: str, keywords: Set[str],
            allow_upper=False,
            allow_first_letter_upper=False,
            allowed_letters_before_keyword_regex: str = "", allowed_letters_after_keyword_regex: str = ""
    ) -> None:
        hyperscan_database = hyperscan.Database()
        loguru.logger.debug(f"Staring compilation for feature {name}")
        hyperscan_database.compile(
            expressions=[bytes(cls.compute_regex_string(
                keyword,
                allow_upper=allow_upper,
                allow_first_letter_upper=allow_first_letter_upper,
                allowed_letters_before_keyword_regex=allowed_letters_before_keyword_regex,
                allowed_letters_after_keyword_regex=allowed_letters_after_keyword_regex,
                hyperscan_compatible=True
            ), 'utf-8') for keyword in keywords],
            ids=range(len(keywords)),
            elements=len(keywords)
        )
        loguru.logger.debug(f"Finished compilation for feature {name}")
        db_path = Path(mim_ocr_cfg.features.keyword_features_hyperscan_databases_directory_path)/name
        with open(db_path, "wb") as db_file:
            db_file.write(hyperscan.dumpb(hyperscan_database))

    @classmethod
    def compute_regex_string(
        cls, keyword, allow_upper: bool = False, allow_first_letter_upper: bool = False,
        allowed_letters_before_keyword_regex: str = "", allowed_letters_after_keyword_regex: str = "",
        hyperscan_compatible: bool = False
    ) -> str:
        regex_string = rf"{re.escape(keyword)}"
        if allow_upper:
            regex_string += rf"|{re.escape(keyword.upper())}"
        if allow_first_letter_upper:
            regex_string += rf"|{re.escape(keyword[0].upper()) + re.escape(keyword[1:])}"
        keyword_inner_regex = rf"({regex_string})"
        before_regex = rf"\s|^{allowed_letters_before_keyword_regex}"
        after_regex = rf"\s|${allowed_letters_after_keyword_regex}"
        if hyperscan_compatible:
            return rf'(?:{before_regex}){keyword_inner_regex}(?:{after_regex})'
        else:
            return rf'(?:{before_regex}){keyword_inner_regex}(?={after_regex})'

    @staticmethod
    def on_match(keyword_index: int, fromm: int, to: int, flags: int,
                 context: Optional[Dict[str, Any]] = None) -> Optional[bool]:
        """See https://python-hyperscan.readthedocs.io/en/latest/usage/"""

        context['matching_ids'].add(keyword_index)
        return False

    def find_occurrences(self, text: str) -> List[Occurrence]:
        """See https://python-hyperscan.readthedocs.io/en/latest/usage/"""
        occurrences: List[Occurrence] = []
        matching_ids = set()
        self.hyperscan_database.scan(bytes(text, 'utf-8'), match_event_handler=self.on_match, context={
            'matching_ids': matching_ids,
        })
        for keyword_index in matching_ids:
            keyword = self.keywords[keyword_index]
            regex_string = self.compute_regex_string(
                keyword,
                allow_upper=self.allow_upper,
                allow_first_letter_upper=self.allow_first_letter_upper,
                allowed_letters_before_keyword_regex=self.allowed_letters_before_keyword_regex,
                allowed_letters_after_keyword_regex=self.allowed_letters_after_keyword_regex,
            )
            for m in re.finditer(regex_string, text):
                if m.groups():
                    occurrences.append(
                        Occurrence(text, self.name, m.span(1)[0], m.span(1)[1], m.group(1), self.priority))
        return occurrences


class DataFrameBasedKeywordFeature(KeywordFeature):
    def __init__(self, name: str, keywords_df: pd.DataFrame, keyword_columns: Optional[List[str]], priority: int = 5,
                 allow_upper: bool = False, allow_first_letter_upper: bool = False,
                 allowed_letters_before_keyword_regex: str = "", allowed_letters_after_keyword_regex: str = "",
                 allow_recompute: bool = False):
        keywords = keywords_df.values.tolist()
        keyword_list = [item.strip() for sublist in keywords for item in sublist if not isnan(item)]

        super().__init__(name, keyword_list, priority, allow_upper, allow_first_letter_upper,
                         allowed_letters_before_keyword_regex=allowed_letters_before_keyword_regex,
                         allowed_letters_after_keyword_regex=allowed_letters_after_keyword_regex,
                         allow_recompute=allow_recompute)


class ExcelKeywordFeature(DataFrameBasedKeywordFeature):
    def __init__(self, name: str, excel_file_path: str, keyword_columns: Optional[List[str]], priority: int = 5,
                 allow_upper: bool = False,
                 allowed_letters_before_keyword_regex: str = "", allowed_letters_after_keyword_regex: str = "",
                 allow_recompute: bool = False):
        keywords_df = pd.read_excel(excel_file_path, usecols=keyword_columns)
        super().__init__(name, keywords_df, keyword_columns, priority, allow_upper,
                         allowed_letters_before_keyword_regex=allowed_letters_before_keyword_regex,
                         allowed_letters_after_keyword_regex=allowed_letters_after_keyword_regex,
                         allow_recompute=allow_recompute)


class CSVKeywordFeature(DataFrameBasedKeywordFeature):
    def __init__(self, name: str, csv_file_path: str, keyword_columns: Optional[List[str]], priority: int = 5,
                 allow_upper: bool = False, allow_first_letter_upper: bool = False,
                 allowed_letters_before_keyword_regex: str = "", allowed_letters_after_keyword_regex: str = "",
                 allow_recompute: bool = False):
        keywords_df = pd.read_csv(csv_file_path, usecols=keyword_columns)
        super().__init__(name, keywords_df, keyword_columns, priority, allow_upper, allow_first_letter_upper,
                         allowed_letters_before_keyword_regex=allowed_letters_before_keyword_regex,
                         allowed_letters_after_keyword_regex=allowed_letters_after_keyword_regex,
                         allow_recompute=allow_recompute)



from typing import Tuple

import cv2
import numpy as np
from google.cloud import vision
from google.cloud.vision_v1 import Word, BoundingPoly

from mim_ocr.backends import OCRBackend
from mim_ocr.data_model import Box
from mim_ocr.data_model.box import BoxType


class GCPBackend(OCRBackend):
    """based on https://github.com/SoloSynth1/gcp-vision-ocr/blob/master/vision.py"""

    def run_ocr_to_box(self, img: np.ndarray, *args, **kwargs) -> Box:

        img_str = cv2.imencode('.png', img)[1].tostring()

        client = vision.ImageAnnotatorClient()

        image = vision.Image(content=img_str)
        response = client.document_text_detection(image=image)
        document = response.full_text_annotation

        # https://cloud.google.com/vision/docs/fulltext-annotations
        root_box = Box.create_root_box()
        for page in document.pages:
            page_box = Box(left=0, top=0, right=page.width, bottom=page.height,
                           conf=None, text=None, box_type=BoxType.GCP_DOCUMENT)
            Box.add_child(root_box, page_box)

            for block in page.blocks:
                # Following block types are supported by Google OCR, but are not parsed by MIM OCR:
                # BlockType.BARCODE
                # BlockType.PICTURE
                # BlockType.RULER
                # BlockType.TABLE
                # BlockType.UNKNOWN
                if block.block_type == block.BlockType.TABLE:
                    box_type = BoxType.GCP_BLOCK_TABLE  # TODO: check whether text extraction for tables works.
                elif block.BlockType.TEXT:
                    box_type = BoxType.GCP_BLOCK_TEXT
                else:
                    continue

                (left, top, right, bottom) = self._get_rectangle_from_bounding_box(block.bounding_box)
                block_box = Box(left=left, top=top, right=right, bottom=bottom,
                                conf=block.confidence*100, text=None, box_type=box_type)
                Box.add_child(page_box, block_box)

                for paragraph in block.paragraphs:
                    (left, top, right, bottom) = self._get_rectangle_from_bounding_box(paragraph.bounding_box)
                    paragraph_box = Box(left=left, top=top, right=right, bottom=bottom,
                                        conf=block.confidence*100, text=None,
                                        box_type=BoxType.GCP_BLOCK_PARAGRAPH)
                    Box.add_child(block_box, paragraph_box)

                    for word in paragraph.words:
                        (left, top, right, bottom) = self._get_rectangle_from_bounding_box(word.bounding_box)
                        text = self._get_text_for_word(word)
                        word_box = Box(left=left, top=top, right=right, bottom=bottom,
                                       conf=block.confidence*100, text=text,
                                       box_type=BoxType.GCP_BLOCK_WORD)
                        Box.add_child(paragraph_box, word_box)
        return root_box

    @staticmethod
    def _get_text_for_word(word: Word) -> str:
        return "".join([symbol.text for symbol in word.symbols])

    @staticmethod
    def _get_rectangle_from_bounding_box(bounding_box: BoundingPoly) -> Tuple[int, int, int, int]:
        y_values = [vertice.y for vertice in bounding_box.vertices]
        x_values = [vertice.x for vertice in bounding_box.vertices]

        return min(x_values), min(y_values), max(x_values), max(y_values)



import itertools
from typing import List, Optional

from mim_ocr.data_model import Box
from mim_ocr.heuristics import Feature, Occurrence


class TextLineAnalysis:
    def __init__(self, original_text: str, features: Optional[List[Feature]] = None):
        if features is None:
            features = []

        self.original_text = original_text
        self.feature_occurrences = list(
            itertools.chain.from_iterable([feature.find_occurrences(self.original_text) for feature in features]))

    def __repr__(self):
        return f"{self.original_text}: {self.feature_occurrences}"

    __str__ = __repr__


def map_features_onto_words(tesseract_line_box: Box, feature_occurrences: List[Occurrence]) -> None:
    """As we calculated feature for text lines, we need to map them onto specific words e.g. for visualization"""
    position = 0
    for word_box in tesseract_line_box.children:
        if word_box.text:
            word_length = len(word_box.text)
            for occurrence in feature_occurrences:
                if position <= occurrence.start < position + word_length:
                    word_box.additional_data = {"feature": occurrence.feature.__name__.replace("Feature", "")}
                    break
            if not word_box.additional_data:
                word_box.additional_data = {"feature": ""}
        position += word_length + 1



from typing import Tuple

import boto3
import numpy as np
import cv2

from mim_ocr.backends import OCRBackend
from mim_ocr.data_model import Box
from mim_ocr.data_model.box import BoxType


class AwsTextractBackend(OCRBackend):

    def run_ocr_to_box(self, img: np.ndarray, *args, **kwargs) -> Box:
        client = boto3.client('textract')
        img_height, img_width = img.shape[0], img.shape[1]
        img2 = cv2.imencode('.png', img)[1]
        response = client.detect_document_text(Document={'Bytes': img2.tobytes()})
        return self.response_to_box(response, img_height, img_width)

    def response_to_box(self, response: dict, img_height: int, img_width: int) -> Box:
        root_box = Box.create_root_box()
        root_box.additional_data = {'DocumentMetadata': response['DocumentMetadata']}

        pages = [b for b in response['Blocks'] if b['BlockType'] == 'PAGE']
        lines = [b for b in response['Blocks'] if b['BlockType'] == 'LINE']
        words = [b for b in response['Blocks'] if b['BlockType'] == 'WORD']

        for page in pages:
            page_box = Box(conf=None, text=None, box_type=BoxType.AWS_BLOCK_PAGE)
            Box.add_child(root_box, page_box)

            child_lines = [line for line in lines if line['Id'] in page['Relationships'][0]['Ids']]
            for line in child_lines:
                left, top, right, bottom = self.geometry_to_boundingbox(line, img_height, img_width)
                confidence = line['Confidence']
                line_box = Box(left=left, top=top, right=right, bottom=bottom,
                               conf=confidence, text=None, box_type=BoxType.AWS_BLOCK_LINE)
                Box.add_child(page_box, line_box)

                child_words = [word for word in words if word['Id'] in line['Relationships'][0]['Ids']]
                for word in child_words:
                    confidence = word['Confidence']
                    text = word['Text']
                    left, top, right, bottom = self.geometry_to_boundingbox(word, img_height, img_width)
                    word_box = Box(left=left, top=top, right=right, bottom=bottom,
                                   conf=confidence, text=text, box_type=BoxType.AWS_BLOCK_WORD)
                    Box.add_child(line_box, word_box)
        return root_box

    @staticmethod
    def geometry_to_boundingbox(block: dict, img_height: int, img_width: int) -> Tuple[int, int, int, int]:
        top = img_height * block['Geometry']['BoundingBox']['Top']
        left = img_width * block['Geometry']['BoundingBox']['Left']
        width = img_width * block['Geometry']['BoundingBox']['Width']
        height = img_height * block['Geometry']['BoundingBox']['Height']
        right = left + width
        bottom = top + height

        return int(left), int(top), int(right), int(bottom)



import argparse
import dataclasses
import multiprocessing
import os
from contextlib import closing
from multiprocessing import Pool
from multiprocessing.pool import ApplyResult
from pathlib import Path
from typing import Optional, List, Callable

from loguru import logger
from tqdm import tqdm

from mim_ocr.backends import OCRBackend
from mim_ocr.heuristics import Feature
from mim_ocr.pipeline.pipeline import run_pipeline_and_save_results_to_file, RunPipelineAndSaveResultToFileInput


@dataclasses.dataclass
class RunPipelineAndSaveDataframeInput:
    image_input_path: str
    out_dir: Optional[str]
    prep_dir: Optional[str]
    nr_proc: int
    backend: Optional[OCRBackend]
    input_box_path: Optional[str]
    preprocessing_transformations: List[Callable] = dataclasses.field(default_factory=lambda: [])
    features: List[Feature] = dataclasses.field(default_factory=lambda: [])

    input_img_filepaths: List[Optional[Path]] = dataclasses.field(default_factory=lambda: [])
    input_box_filepaths: List[Optional[Path]] = dataclasses.field(default_factory=lambda: [])
    preprocessed_image_paths: List[Optional[Path]] = dataclasses.field(default_factory=lambda: [])
    output_filepaths: List[Optional[Path]] = dataclasses.field(default_factory=lambda: [])

    batch_size: int = 1

    def validate(self):
        if self.image_input_path:
            if not os.path.isdir(self.image_input_path):
                raise ValueError("input_img_dir is not a valid directory path.")

            if self.out_dir is not None and not os.path.isdir(self.out_dir):
                raise ValueError("out_dir is not a valid directory path.")

            if self.prep_dir is not None and not os.path.isdir(self.prep_dir):
                raise ValueError("prep_dir is not a valid directory path.")

        if self.input_box_path:
            if not os.path.isdir(self.input_box_path):
                raise ValueError("input_img_dir is not a valid directory path.")

    def calculate_path_lists(self) -> None:
        if self.image_input_path:
            filenames = [f for f in os.listdir(self.image_input_path) if
                         os.path.isfile(os.path.join(self.image_input_path, f))]
            input_filepaths = [Path(os.path.join(self.image_input_path, f)) for f in filenames]
            self.input_img_filepaths = input_filepaths
            self.input_box_filepaths = [None for _ in input_filepaths]

        elif self.input_box_path:
            filenames = [f for f in os.listdir(self.input_box_path) if
                         os.path.isfile(os.path.join(self.input_box_path, f))]
            input_filepaths = [Path(os.path.join(self.input_box_path, f)) for f in filenames]
            self.input_box_filepaths = input_filepaths
            self.input_img_filepaths = [None for _ in input_filepaths]

        if self.out_dir is None:
            self.output_filepaths = [None for _ in input_filepaths]
        else:
            self.output_filepaths = [Path(self.out_dir, f + ".json") for f in filenames]

        if self.prep_dir is None:
            self.preprocessed_image_paths = [None for _ in input_filepaths]
        else:
            self.preprocessed_image_paths = [Path(self.out_dir, f) for f in filenames]

    def get_single_pipeline_input(self, i: int, j: int) -> List[RunPipelineAndSaveResultToFileInput]:
        return [RunPipelineAndSaveResultToFileInput(
            output_path=self.output_filepaths[k],
            image_input_path=self.input_img_filepaths[k],
            preprocessing_transformations=self.preprocessing_transformations,
            preprocessed_image_path=self.preprocessed_image_paths[k],
            backend=self.backend,
            box_input_path=self.input_box_filepaths[k],
            features=self.features,
        ) for k in range(i, j)]


def batch_run_pipeline_and_save_dataframe_for_dirs(args: RunPipelineAndSaveDataframeInput) -> None:
    """Run pipeline for all files in input directory on multiple processors.
    Save result datasheets into output directory."""

    args.validate()
    args.calculate_path_lists()

    logger.info(f"Running pipeline on {len(args.input_img_filepaths)} images.")

    if args.nr_proc == 1:
        # some elements of the pipeline, like NER_FEATURE do not run in multiprocessing environement.
        # Disabling multiprocessing for 1 CPU enables to run them.
        for i, _ in enumerate(tqdm(args.input_img_filepaths)):
            pipeline_input = args.get_single_pipeline_input(i, i+1)
            options_dict = {
                'job_info': f"filepath: {pipeline_input[0].image_input_path or pipeline_input[0].box_input_path} "
                            f"({args.batch_size} files)",
                'suppress_exceptions': True,
            }
            run_pipeline_and_save_results_to_file(pipeline_input, **options_dict)

    else:
        with closing(Pool(args.nr_proc)) as pool:
            funclist: List[ApplyResult] = []
            i = 0
            j = min(i + args.batch_size, len(args.input_img_filepaths))
            while i < j:
                pipeline_input = args.get_single_pipeline_input(i, j)
                options_dict = {
                    'job_info': f"filepath: {pipeline_input[0].image_input_path or pipeline_input[0].box_input_path} "
                                f"({args.batch_size} files)",
                    'suppress_exceptions': True,
                }

                f = pool.apply_async(run_pipeline_and_save_results_to_file,
                                     (pipeline_input,),
                                     options_dict)
                funclist.append(f)
                i = j
                j = min(i + args.batch_size, len(args.input_img_filepaths))

            for k, f in enumerate(tqdm(funclist)):
                try:
                    f.get(timeout=10000*args.batch_size)
                except multiprocessing.context.TimeoutError:
                    logger.error(
                        f"Mutiprocessing timeout error "
                        f"on input "
                        f"{args.input_img_filepaths[args.batch_size] or args.input_box_filepaths[args.batch_size]} "
                        f"({args.batch_size} files).")


class OcrBatchProcessingArgumentParser(argparse.ArgumentParser):
    def __init__(self):
        super().__init__(description='batch run OCR pipeline ans save results do dataframes')

        self.add_argument('--input_img_dir', type=str, help='Directory with input images.')
        self.add_argument('--input_box_dir', type=str, help='Directory with input images.')
        self.add_argument('--out_dir', type=str, help='Directory to store output datasheets.', default=None)
        self.add_argument('--prep_dir', type=str, help='Directory to store preprocessed images.', default=None)
        self.add_argument('--nr_proc', type=int, help='number of processors to use.', default=1)
        self.add_argument('--logfile', type=Path, help='path to logfile')
        self.add_argument('--backend', type=str, help='Backend name, e.g TesseractBacked, EasyOCRBackend',
                          default=None)
        self.add_argument('--batch_size', type=int, help='Number of files processed with one worker')
        self.add_argument('--reorient', action=argparse.BooleanOptionalAction)
        self.add_argument('--deskew', action=argparse.BooleanOptionalAction)
        self.add_argument('--features', nargs='+', help='List of features to find')

    def parse_args(self, *args, **kwargs):
        parser_args = super().parse_args(*args, **kwargs)

        if parser_args.logfile:
            logger.remove()
            logger.add(parser_args.logfile, mode="w")

        if parser_args.out_dir and not (parser_args.backend or parser_args.input_box_dir):
            raise ValueError("You need to provide backend together with out_ocr_dir.")

        if not (parser_args.out_dir or parser_args.prep_dir):
            raise ValueError("Nothing to do, no output dirs provided.")

        return parser_args



from .heuristic_analysis import heuristic_examine_box_lines, find_heuristic_features
from .feature import Occurrence, Feature
from .regex_feature import RegexFeature
from .keyword_feature import KeywordFeature, ExcelKeywordFeature, CSVKeywordFeature
from .basic_features import NUMBER_FEATURE, PHONE_NUMBER_FEATURE, DATE_FEATURE, PESEL_FEATURE



import os
import requests
import ipykernel
import re
from notebook.notebookapp import list_running_servers
from pathlib import Path


def get_current_notebook_path() -> Path:
    """
    Code was copied (and slightly modified) from: https://github.com/jupyterhub/jupyterhub/issues/1718
    Popular (and much simplier_ solutions (from StackOverflow or github) fail when:
      * code was moved to python libs (from notebooks) e.g. solutions based on "__file__",
      * notebook cells were run without break (changing kernel status to idle) e.g. solutions based on javascript.
    This solution is a hack and can stop working on newer versions of used libraries.
    This solution works only on jupyter with token based authorization (e.g. in jupyterhub).

    It also does not work when running jupyter notebooks from command line e.g. via jupyter nbconvert.
    TODO detect running from command line (unfortunatelly it sounds as difficult as writing this function
    all SO answers does not support all cases):
    https://stackoverflow.com/questions/15411967/how-can-i-check-if-code-is-executed-in-the-ipython-notebook
    """

    # proper way of getting token in jupyterhub
    TOKEN = os.environ.get('JUPYTERHUB_API_TOKEN', '')

    # proper way of getting token in (single user) jupyter notebooks with token authorization
    running_server_props = next(list_running_servers())
    if not TOKEN:
        TOKEN = running_server_props['token']
    if not TOKEN:
        raise NotImplementedError('Jupyter token not found. Getting current notebook path is not supported '
                                  'for jupyter instances without token authorization')

    base_url = running_server_props['url']
    r = requests.get(
        url=base_url + 'api/sessions',
        headers={'Authorization': 'token {}'.format(TOKEN)})

    r.raise_for_status()
    response = r.json()

    kernel_id = re.search('kernel-(.*).json', ipykernel.connect.get_connection_file()).group(1)
    notebook_path = {r['kernel']['id']: r['notebook']['path'] for r in response}[kernel_id]

    return Path(notebook_path)


def get_directory_for_notebook_images(initial_cwd: Path, makedirs: bool = True) -> Path:
    """
       Get directory path for storing images from current jupyter notebook.
       Args:
           makedirs (bool): if true directory will be created
    """
    dirname = get_current_notebook_path().name + ".images"
    dirpath = Path(initial_cwd) / dirname
    if makedirs:
        os.makedirs(dirpath, exist_ok=True)
    return dirpath



import os
import sys
import torch
from loguru import logger

# import NER resources, if present
ner_resources_path = os.environ.get("NER_RESOURCES_PATH")
if ner_resources_path:
    if os.path.exists(ner_resources_path):
        sys.path.append(os.path.abspath(ner_resources_path))
        from core.polrobertner import PolRobertaNer
        from core.model.xlmr_for_token_classification import XLMRForTokenClassification


class PolRobertaNerOnDevice(PolRobertaNer):
    # this class is based on PolRobertaNer, but allows to specify whether the computations should be done on cpu or cuda

    def __init__(self, model_path: str, roberta_embeddings_path: str, device: str = None):
        # code of this constructor was copied from the parent class PolRobertaNer
        # and modified, to allow specifying the device on which, the model is run ("cpu" or "cuda")
        if not os.path.exists(model_path):
            raise ValueError("Model not found on path '%s'" % model_path)

        if not os.path.exists(roberta_embeddings_path):
            raise ValueError("RoBERTa language model not found on path '%s'" % roberta_embeddings_path)

        if not device:
            if torch.cuda.is_available():
                device = "cuda"
            else:
                device = "cpu"
        logger.info(f'NER computations using {device} device')

        self.label_list = PolRobertaNer.load_labels(os.path.join(model_path, 'labels.txt'))
        model = XLMRForTokenClassification(pretrained_path=roberta_embeddings_path,
                                           n_labels=len(self.label_list) + 1,
                                           hidden_size=768 if 'base' in roberta_embeddings_path else 1024,
                                           device=device)
        with open(os.path.join(model_path, 'model.pt'), 'rb') as model_file:
            state_dict = torch.load(model_file, map_location=torch.device(device))
        model.load_state_dict(state_dict)
        model.eval()
        self.model = model



from pathlib import Path
from typing import Tuple

import cv2
import numpy as np
from pytesseract import pytesseract, Output

MINIMAL_REORIENTATION_CONFIDENCE = 3.0


def reorient_cv2(image: np.ndarray, image_path: Path) -> Tuple[np.ndarray, int]:
    # We need to add path due to strange tesseract errors.
    # https://stackoverflow.com/questions/67018785/why-pytesseract-cant-handle-osd-mode
    # if image_path is None:
    #     im = Image.fromarray(image)
    #     temp = tempfile.NamedTemporaryFile()
    #     temp.write(im.tobytes())
    #     image_path = temp.name
    # else:
    #     image_path = str(image_path)

    osd = pytesseract.image_to_osd(str(image_path), output_type=Output.DICT)
    rotation = osd['rotate']
    confidence = osd['orientation_conf']
    if confidence < MINIMAL_REORIENTATION_CONFIDENCE:
        rotation = 0

    return _reorient_image(image, rotation), rotation


def _reorient_image(image: np.ndarray, angle: int):
    if angle == 0:
        return image
    elif angle == 90:
        return cv2.rotate(image, cv2.cv2.ROTATE_90_CLOCKWISE)
    elif angle == 180:
        return cv2.rotate(image, cv2.cv2.ROTATE_180)
    elif angle == 270:
        return cv2.rotate(image, cv2.cv2.ROTATE_90_COUNTERCLOCKWISE)
    raise ValueError("Incorrect rotation. Must be multiple of 90.")



import os
import re
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import List, Optional

# Dictionary mapping features names to their priority.
# The box read from a file may already have some features' occurrences recognized.
# When computing some other features on such box, we use this dictionary to check,
# if the feature occurrence found now for a given word, should overwrite feature occurrence found previously.
FEATURES_PRIORITIES_DICTIONARY = {"Number": 5, "PhoneNumber": 7, "Date": 8, "PESEL": 10}

if os.environ.get("NER_RESOURCES_PATH"):
    FEATURES_PRIORITIES_DICTIONARY["NERFeature"] = 2


@dataclass(eq=True, frozen=True)
class Occurrence:
    full_text: str
    feature_name: str
    start: int
    end: int
    matched_text: str
    priority: int
    additional_info: Optional[str] = None

    @staticmethod
    def get_occurrences_from_regex(regex: re.Pattern, text: str, feature_name: str, feature_priority: int) \
            -> List['Occurrence']:

        occurrences: List[Occurrence] = []
        for m in re.finditer(regex, text):
            if m.groups():
                occurrences.append(
                    Occurrence(text, feature_name, m.span(1)[0], m.span(1)[1], m.group(1), feature_priority))
        return occurrences


class Feature(ABC):

    def __init__(self, name: str, priority: int = 5):
        """
        Args:
            name: name of the feature
            priority:  priority will be used when two features overlap and one of features needs to be
                       chosen or presented.
                       higher number (like 9) means detailed features (like PESEL).
                       lower number (like 0) means general features (like Number).
        """

        self.name = name
        self.priority = priority

    @abstractmethod
    def find_occurrences(self, text: str) -> List[Occurrence]:
        pass

    @classmethod
    def get_feature_priority_by_name(cls, name: str) -> int:

        # remove "<-", "-" and "->" suffixes
        name = name.rstrip("<->")

        if name in FEATURES_PRIORITIES_DICTIONARY:
            return FEATURES_PRIORITIES_DICTIONARY[name]

        # NERFeature occurrences have different names, depending on the category recognized by NER,
        # but everytime it has "NERFeature" suffix.
        elif name.endswith("NERFeature"):
            return FEATURES_PRIORITIES_DICTIONARY["NERFeature"]

        raise ValueError(f"Cannot determine priority for {name} feature occurrence.")



import re
from typing import List

from .feature import Feature, Occurrence


class RegexFeature(Feature):
    def __init__(self, name: str, core_regex: str, before_regex: str = "", after_regex: str = "", priority: int = 5):
        """
        Args:
            name:
            core_regex: regex that will be matched in feature occurrence
            before_regex: regex that needs to match text before feature occurrence
            after_regex: regex that needs to match text after feature occurrence
            priority:
        """

        super().__init__(name, priority=priority)
        self.regex = re.compile(rf'(?:{before_regex})({core_regex})(?={after_regex})')
        # "({core_regex})(?={after_regex})" means, that pattern of core_regex
        # has to be followed by pattern of after_regex, to be matched,
        # but after_regex pattern is not included in the match

    def find_occurrences(self, text: str) -> List[Occurrence]:
        return Occurrence.get_occurrences_from_regex(self.regex, text, self.name, self.priority)



import traceback
import PIL
from pytesseract import TesseractError

from loguru import logger


# Context Manager that helps to catch and log error when running batch jobs
class SmoothOCRJobRunContext:
    """Context Manager that handles logging and exception handling during single tesseract execution.
       Exceptions raised by jobs can be suppressed.
    """

    def __init__(self, job_info: str = "", suppress_exceptions: bool = True):
        self.job_info = job_info
        self.suppress_exceptions = suppress_exceptions

    def __enter__(self):
        logger.debug(f"Running {self.job_info}.")

        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if not exc_type:
            return True
        if exc_type in (
                KeyboardInterrupt, BrokenPipeError, ImportError, ChildProcessError, BlockingIOError,
                MemoryError,
                SyntaxError, ModuleNotFoundError, SystemError):
            logger.error(f"Error blocking execution for {self.job_info}")
            logger.debug("".join(traceback.format_exception(exc_type, exc_value, exc_traceback)))
            return False

        if exc_type in (TesseractError, ValueError, PIL.Image.DecompressionBombError):
            logger.info(f"Known Error for {self.job_info}")
            logger.info(f"ERROR VALUE: {exc_value}")
            logger.debug("".join(traceback.format_exception(exc_type, exc_value, exc_traceback)))
            return self.suppress_exceptions

        logger.info(f"Unknown Error blocking execution for {self.job_info}")
        logger.debug("".join(traceback.format_exception(exc_type, exc_value, exc_traceback)))
        return self.suppress_exceptions



from pathlib import Path
from typing import List

import cv2
import numpy as np
import pdf2image
from PIL import Image


def open_image(path: Path) -> np.ndarray:
    """ Opens image from path as cv2 matrix.
        Using PIL.Image as a proxy here gives us support for all image formats supported by tesseract:
        https://github.com/tesseract-ocr/tessdoc/blob/main/InputFormats.md
    """
    with Image.open(path) as pil_image:
        return pil_image_to_cv2(pil_image)


def pil_image_to_cv2(pil_image: Image) -> np.ndarray:
    # based on https://stackoverflow.com/questions/14134892/convert-image-from-pil-to-opencv-format
    return cv2.cvtColor(np.array(pil_image.convert('RGB')), cv2.COLOR_RGB2BGR)


def open_pdf(path: Path) -> List[np.ndarray]:
    return [pil_image_to_cv2(pil_image)
            for pil_image in pdf2image.convert_from_path(path)]


def write_image(img: np.ndarray, path: Path):
    """ based on: https://stackoverflow.com/a/11335576/3178243
    """
    cv2.imwrite(str(path), img, [int(cv2.IMWRITE_JPEG_QUALITY), 90])



import os

import torch.cuda

if os.environ.get("NER_RESOURCES_PATH"):
    from .NER_feature import NERFeature


if os.environ.get("NER_RESOURCES_PATH"):
    NER_FEATURE = NERFeature(device='cpu')

    if torch.cuda.is_available():
        NER_FEATURE_CUDA = NERFeature(device='cuda')



from abc import ABC, abstractmethod

import numpy as np
import pandas as pd

from mim_ocr.data_model import Box
from mim_ocr.utils.class_utils import get_subclass_by_name


class OCRBackend(ABC):

    @abstractmethod
    def run_ocr_to_box(self, img: np.ndarray, *args, **kwargs) -> Box:
        pass

    def run_ocr_to_dataframe(self, img: np.ndarray, *args, **kwargs) -> pd.DataFrame:
        if img is None:
            raise ValueError("Input image cannot be None")

        return self.run_ocr_to_box(img, *args, **kwargs).to_dataframe()

    @staticmethod
    def get_by_name(backend_name: str) -> 'OCRBackend':
        return get_subclass_by_name(OCRBackend, backend_name)()


class OCRBackendException(Exception):
    pass



from itertools import chain
from typing import Type, List


def get_subclass_by_name(base_class: Type, name: str):
    for subclass in _get_all_subclasses(base_class):
        if subclass.__name__ == name:
            return subclass
    raise KeyError("Subclass of this name not found.")


def _get_all_subclasses(base_class: Type) -> List[Type]:
    return list(chain.from_iterable([_get_all_subclasses(sub_class) for sub_class in base_class.__subclasses__()])) \
           + [base_class]



import functools
import os
from omegaconf import OmegaConf

mim_ocr_config_path = os.environ.get('MIM_OCR_CONFIG_PATH')

if mim_ocr_config_path:
    mim_ocr_cfg = OmegaConf.load(mim_ocr_config_path)
else:
    mim_ocr_cfg = None


def mim_ocr_cfg_required(func):
    """Decorator for checking that mim_ocr_cfg is present"""
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        check_mim_ocr_cfg_exists()
        return func(*args, **kwargs)

    return wrapper


def check_mim_ocr_cfg_exists():
    if not mim_ocr_cfg:
        raise ValueError("MIM_OCR_CONFIG_PATH environmental variable not set. You need to create a YAML config file, "
                         "and specify path in this variable. Example file can be found in "
                         "config/test_mim_ocr_conf.yaml.")



import re
import sys
import os
from typing import List
from ...heuristics.feature import Feature, Occurrence

# import NER resources, if present
ner_resources_path = os.environ.get("NER_RESOURCES_PATH")
if ner_resources_path:
    if os.path.exists(ner_resources_path):
        sys.path.append(os.path.abspath(ner_resources_path))
        from mim_ocr.heuristics.pol_roberta_ner_on_device import PolRobertaNerOnDevice


class NERFeature(Feature):
    def __init__(self, device='cpu'):
        if ner_resources_path:
            if os.path.exists(ner_resources_path):
                super().__init__(name="NERFeature", priority=2)

                pretrained_model_path = os.path.join(ner_resources_path, "models", "ner_pretrained_model")
                # The model was trained as described in:
                # https://github.com/mczuk/xlm-roberta-ner#train-the-model-from-scratch

                roberta_base_model_path = os.path.join(ner_resources_path, "models", "roberta_base_fairseq")
                self.ner_model = PolRobertaNerOnDevice(pretrained_model_path, roberta_base_model_path, device=device)

            else:
                raise ValueError("Path to NER resources isn't specified correctly, check README file")
        else:
            raise ValueError("Path to NER resources isn't specified correctly, check README file")

    def find_occurrences(self, text: str) -> List[Occurrence]:
        occurrences = []

        if text.strip():  # check if input string isn't empty or whitespace
            words_and_spaces = list(filter(None, re.split(r'(\s)', text)))
            labels = [None] * len(words_and_spaces)
            words = [re.split(r'\s', text)]
            words_labels = self.ner_model.process(words)
            # process() takes a list of list of words as an argument
            # and returns a corresponding list of list of labels.
            # Each list of words constitutes a context for its elements.
            # Here, we analyze only one string at a time,
            # so process() takes a list containing one list of words as an argument,
            # and returns a list containing one list of labels.

            j = 0
            for i in range(len(words_and_spaces)):
                if not re.match(r"\s", words_and_spaces[i]):
                    labels[i] = words_labels[0][j]
                    j += 1

            word_start = 0
            for word, label in zip(words_and_spaces, labels):
                word_end = word_start + len(word)
                if label and label != "O":
                    # label "O" means that the model didn't find any relevant category for a word
                    feature_label = label + "-" + self.name
                    occurrence = Occurrence(full_text=text, feature_name=feature_label, start=word_start,
                                            end=word_end, matched_text=word, priority=self.priority)
                    occurrences.append(occurrence)
                word_start = word_end

        return occurrences



import dataclasses
from pathlib import Path
from typing import List, Callable, Optional, Tuple

import cv2
import numpy as np

from mim_ocr.backends import OCRBackend
from mim_ocr.data_model import Box
from mim_ocr.exceptions.smooth_job_context import SmoothOCRJobRunContext
from mim_ocr.heuristics import Feature, heuristic_examine_box_lines
from mim_ocr.image import open_image


def run_ocr_pipeline_on_file(input_path: Path, preprocessing_transformations: List[Callable],
                             backend: Optional[OCRBackend]) -> Tuple[np.ndarray, Optional[Box]]:
    img = open_image(input_path)
    metadata = {'path': input_path}

    for t in preprocessing_transformations:
        img = t(img, input_path, metadata)

    if backend is not None:
        return img, backend.run_ocr_to_box(img)
    else:
        return img, None


@dataclasses.dataclass
class RunPipelineAndSaveResultToFileInput:
    output_path: Optional[Path]
    image_input_path: Optional[Path] = None
    preprocessed_image_path: Optional[Path] = None
    preprocessing_transformations: Optional[List[Callable]] = dataclasses.field(default_factory=lambda: [])
    backend: Optional[OCRBackend] = None
    box_input_path: Optional[Path] = None
    features: List[Feature] = dataclasses.field(default_factory=lambda: [])

    def validate(self):
        if not (self.image_input_path or self.box_input_path):
            raise ValueError("You have to provide image_input_path or box_input_path")
        if self.image_input_path and not self.backend:
            raise ValueError("You have to provide backend")
        if self.box_input_path and (self.image_input_path
                                    or self.preprocessed_image_path
                                    or self.preprocessing_transformations
                                    or self.backend):
            raise ValueError("You cannot use image_input_path or preprocessed_image_path "
                             "or preprocessing_transformations or backend when using precomputed boxes.")
        if self.box_input_path and not self.output_path:
            raise ValueError("You need to specify output path when using precomputed boxes")
        if self.box_input_path and not self.features:
            raise ValueError("No features defined, nothing to do.")

    def read_box(self) -> Box:
        if str(self.box_input_path).endswith('.json'):
            return Box.from_json_file(self.box_input_path)
        if str(self.box_input_path).endswith('.xlsx'):
            return Box.from_excel(self.box_input_path)
        if str(self.box_input_path).endswith('.csv'):
            return Box.from_csv(self.box_input_path)
        raise ValueError("Unrecognized Box file format.")


def run_pipeline_and_save_results_to_file(args_list: List[RunPipelineAndSaveResultToFileInput],
                                          suppress_exceptions: bool = False,
                                          job_info: str = "") -> None:
    """Run full pipeline for an image file.
    Args:
        output_path (Optional[pathlib.Path]): path to output excel file for OCR results
        image_input_path (Optional[pathlib.Path]): path to input image file if pipeline runs on image
        preprocessed_image_path (Optional[pathlib.Path]): path to store preprocessed image
        preprocessing_transformations (List[Callable]): list of image preprocessing transformations
        backend (OCRBackend): if None, then OCR is not performed
        box_input_path (Optional[pathlib.Path]): path to input box file if pipeline runs on box images
        features (Optional[List[Feature]]): Space-separated list of names of Feateres to search in OCR results, example:
                                             NUMBER_FEATURE PHONE_NUMBER_FEATURE NER_FEATURE.
        suppress_exceptions (bool): allows to log and not raise every exception e.g. for batch runs
        job_info (str): additional info for logs
    """
    for args in args_list:
        with SmoothOCRJobRunContext(job_info=job_info, suppress_exceptions=suppress_exceptions):
            args.validate()

            if args.image_input_path:
                img, box = run_ocr_pipeline_on_file(
                    args.image_input_path,
                    args.preprocessing_transformations,
                    args.backend)
                if args.preprocessed_image_path:
                    cv2.imwrite(str(args.preprocessed_image_path), img)

            if args.box_input_path:
                box = args.read_box()

            if args.output_path and args.features:
                heuristic_examine_box_lines(box, features_to_check=args.features)

            if box and args.output_path:
                box.to_json_file(args.output_path)




from typing import Optional

import numpy as np
import pandas as pd
from pytesseract import Output, pytesseract

from .backend import OCRBackend
from mim_ocr.data_model import Box
from mim_ocr.data_model.box import BoxType

DEFAULT_TESSERACT_CONFIG = r'--oem 1 --psm 3 -l pol'


class TesseractBackend(OCRBackend):
    def __init__(self, config: str = DEFAULT_TESSERACT_CONFIG):
        self.config = config

    def run_ocr_to_box(self, img: np.ndarray, *args, config: Optional[str] = None, **kwargs) -> Box:
        if config is None:
            config = self.config
        df = self.run_ocr_to_dataframe(img, config)
        return self.dataframe_to_box(df)

    def run_ocr_to_dataframe(self, img: np.ndarray, config: Optional[str]) -> pd.DataFrame:
        if config is None:
            config = self.config
        if img is None:
            raise ValueError("Input image cannot be None")
        d = pytesseract.image_to_data(img, output_type=Output.DICT, config=config)
        return pd.DataFrame.from_dict(d).astype({'conf': float})

    @staticmethod
    def dataframe_to_box(df: pd.DataFrame) -> Box:
        tree = Box.create_root_box()
        for _, row in df.iterrows():
            right = row.left + row.width
            bottom = row.top + row.height

            new_box = Box(left=row.left, top=row.top, right=right, bottom=bottom,
                          conf=row.conf, text=row.text, box_type=BoxType(row.level))
            tree.add_box_based_on_type(new_box)

        return tree

    @staticmethod
    def calc_average_line_len(box: Box) -> float:
        """Returns average line if found lines. Useful e.g. to assess image rotation."""

        text_lines_boxes = box.get_subboxes(box_type=BoxType.TESSERACT_LINE)
        return sum([text_line_box.width() for text_line_box in text_lines_boxes]) / len(text_lines_boxes)



from pathlib import Path
from typing import Dict

import numpy as np

from mim_ocr.preprocessing import deskew_cv2, reorient_cv2


def no_transform(img: np.ndarray, path: Path, metadata: Dict) -> np.ndarray:
    return img


def reorient(img: np.ndarray, path: Path, metadata: Dict) -> np.ndarray:
    img, orientation = reorient_cv2(img, image_path=path)
    metadata['orientation'] = orientation
    return img


def deskew(img: np.ndarray, path: Path, metadata: Dict) -> np.ndarray:
    img, angle = deskew_cv2(img)
    metadata['angle'] = f"{angle:.2}"
    return img



from typing import Optional, Any, Dict, Tuple

import cv2
import numpy as np
from PIL import Image, ImageDraw, ImageFont

from mim_ocr.data_model.box import Box, BoxType
from mim_ocr.utils import isnan

"""
Script visualizes tesseract results on all files in a directory.
Displays:
 - original image with tesseract boxes
 - image with parsed text and tesseract boxes

usage: tesseract_tests dirname
"""

COLORS = [
    (200, 200, 0),
    (255, 0, 0),
    (0, 255, 0),
    (0, 204, 205),
    (170, 0, 255),
    (255, 127, 0),
    (0, 149, 255),
    (255, 0, 170),
]
COLOR_WHITE = (255, 255, 255)

FONT_COLOR = (0, 0, 0)
FONT_PATH = "/usr/share/fonts/truetype/ubuntu/Ubuntu-B.ttf"
ADDITIONAL_INFO_FONT_SIZE = 12


def visualize_ocr_result(original_image: np.ndarray,
                         box: Box,
                         metadata: Optional[Dict[str, Any]] = None,
                         confidence_threshold: float = 0.0,
                         show_confidence: bool = True,
                         additional_value_to_show: Optional[str] = None,
                         show_only_text_with_additional_value: bool = False,
                         show_outer_boxes: bool = True,
                         only_type: Optional[BoxType] = None,
                         only_original: bool = False,
                         only_ocr_result: bool = False,
                         ) -> np.ndarray:
    """Creates image that visualizes OCR results:
    On left half of the image are shown:
      - original image
      - boxes showing places of OCR
    On right half are shown:
      - boxes showing places of OCR
      - text with OCR result
      - OCR confidence for each text (or additional value if specified)
      - additional information in form of metadata

    Args:
        original_image (np.ndarray): original image (shown as reference).
        box (Box): Box object with OCR results.
        metadata (Optional[Dict[str, Any]]): additional dictionary with additional image legend.
        confidence_threshold (float): between 0.0 to 100.0
        show_confidence (bool): should additional info about word confidence be presented
        additional_value_to_show (Optional[str]): name of key in box items additional data dictionary
                                                  that would be presented instead of confidence.
                                                  This cannot be used together with show_confidence=True.
        show_only_text_with_additional_value (bool): Should shown word be limited to only those
                                                     that have an additional value.
        show_outer_boxes (bool): If should be plotted also bigger boxes (level below the word), default True.
        only_type (Optional[BoxType]): If not None, show only boxes with the specified type, default None.
        only_original: If only original image with bounding boxes should be visualize, defult False
        only_ocr_result: If only OCR parsed image with bounding boxes should be visualize, defult False

    Returns:
        visualization image as ndarray (cv2 image)
    """

    if show_confidence and additional_value_to_show:
        raise ValueError("Cannot show confidence and additional_value.")

    h, w, c = original_image.shape

    # enlarge image to have place to visualize parsed_image
    new_img = np.zeros((h, 2 * w, c), np.uint8)
    new_img[:, :] = COLOR_WHITE
    new_img[0:h, 0:w] = original_image.copy()

    box_type_color_mapping = _generate_box_type_color_mapping(box)

    if show_outer_boxes:
        new_img = add_outer_boxes(
            new_img, box, box_type_color_mapping, confidence_threshold, w,
            only_type=only_type, only_original=only_original)

    if only_original:
        return new_img[:, :w, :]

    # add text on parsed image
    im_parsed = Image.fromarray(new_img)
    draw = ImageDraw.Draw(im_parsed)

    for b in box.preorder_traversal():

        if (b.conf is not None) and (b.conf < confidence_threshold):

            if not show_outer_boxes:
                continue
            if b.conf != -1:
                continue

        if isnan(b.text):
            continue

        if additional_value_to_show:
            if show_only_text_with_additional_value \
                and (additional_value_to_show not in b.additional_data
                     or isnan(b.additional_data[additional_value_to_show])
                     or not b.additional_data[additional_value_to_show]):
                continue

        (x2, y2, w2, h2) = (b.left, b.top, b.width(), b.height())

        fontsize, _ = get_optimal_pillow_font_size(w=w2, h=h2, txt=b.text, d=draw)
        font = ImageFont.truetype(FONT_PATH, fontsize)

        draw.text(xy=(x2 + w, y2), text=b.text, fill=(0, 0, 0), font=font, spacing=0)

        if show_confidence and b.text:
            additional_info_font = ImageFont.truetype(FONT_PATH, ADDITIONAL_INFO_FONT_SIZE)
            draw.text(xy=(x2 + w, y2 - 12), text=f"{b.conf:.1f}", fill=(0, 200, 0), font=additional_info_font,
                      spacing=0)  # parameters obtained by trial and error

        if additional_value_to_show and \
                (additional_value_to_show in b.additional_data) \
                and not isnan(b.additional_data[additional_value_to_show]):
            additional_info_font = ImageFont.truetype(FONT_PATH, ADDITIONAL_INFO_FONT_SIZE)
            draw.text(xy=(x2 + w, y2 - 12), text=f"{b.additional_data[additional_value_to_show]}", fill=(0, 200, 0),
                      font=additional_info_font,
                      spacing=0)  # parameters obtained by trial and error

    # show metadata
    if metadata is not None:
        metadata_fontsize = 40

        metadata_str = "".join(f'{k}: {v}\n' for k, v in metadata.items())

        metadata_font = ImageFont.truetype(FONT_PATH, metadata_fontsize)
        draw.text(xy=(w, 0), text=metadata_str, fill=(0, 200, 0), font=metadata_font, spacing=4)

    if only_ocr_result:
        return np.array(im_parsed)[:, w:, :]

    return np.array(im_parsed)


def add_outer_boxes(
        image: np.ndarray,
        box: Box,
        box_type_color_mapping: Dict[BoxType, Tuple[int, int, int]],
        confidence_threshold: float,
        width: int,
        only_type: Optional[BoxType] = None,
        only_original: bool = False,
) -> np.ndarray:
    for b in box.preorder_traversal():
        if (b.conf is not None) and (b.conf < confidence_threshold):
            if b.conf != -1:
                continue

        if (only_type is not None) and (b.box_type != only_type):
            continue

        # add boxes on the original
        color = box_type_color_mapping[b.box_type]

        image = cv2.rectangle(image, (b.left, b.top), (b.right, b.bottom), color, thickness=2)
        # add boxes on parsed image
        if not only_original:
            image = cv2.rectangle(image, (b.left + width, b.top), (b.right + width, b.bottom), color, thickness=2)

    return image


def get_optimal_pillow_font_size(w: int, h: int, txt: str, d: ImageDraw.Draw) -> Tuple[int, int]:
    """Calculates which font size to use to fill box.
    Uses binary search by font size (with max 20 trials)."""
    font_size = 1
    font_min = 0
    font_max = 0
    for i in range(20):
        new_font = ImageFont.truetype(FONT_PATH, font_size)
        txt_width, txt_height = d.textsize(txt, font=new_font)
        if txt_width == w and txt_height == h:
            return font_size, txt_height
        elif txt_width > w or txt_height > h:
            font_max = font_size
            next_font_size = (font_min + font_size) // 2
        else:
            font_min = font_size
            if font_max:
                next_font_size = (font_max + font_size) // 2
            else:
                next_font_size = font_size * 2

        if next_font_size == font_size:
            return font_size, txt_height
        font_size = next_font_size
    return font_size, txt_height


def compare_ocr_results(img: np.ndarray, box1: Box, box2: Box,
                        confidence_threshold1: float = 0.0,
                        confidence_threshold2: float = 0.0) -> np.ndarray:
    """Creates image that visualizes OCR results:
    On top of the image is shown original image.
    Below on the left half of the image are shown results from the first OCR method:
      - boxes showing places of OCR
      - text with OCR result
    On the right half are shown results from the second OCR method:
      - boxes showing places of OCR
      - text with OCR result

    Args:
        img (np.ndarray): original image (shown as reference).
        box1 (Box): Box object with OCR results.
        box2 (Box): Box object with OCR results.

    Returns:
        visualization image as ndarray (cv2 image)
    """

    h, w, c = img.shape

    # enlarge image to have place to visualize parsed_image
    new_img = np.zeros((2 * h, 2 * w, c), np.uint8)
    new_img[:, :] = COLOR_WHITE
    new_img[0:h, 0:w] = img.copy()

    for box, shift, thres in [(box1, 0, confidence_threshold1), (box2, w, confidence_threshold2)]:
        for b in box.preorder_traversal():

            if (b.conf is not None) and (b.conf < thres):
                continue

            level = b.box_type.value
            col = COLORS[level - 1]

            # add boxes on parsed image
            new_img = cv2.rectangle(new_img, (b.left + shift, b.top + h), (b.right + shift, b.bottom + h), col, 2)

    # add text on parsed image
    im_parsed = Image.fromarray(new_img)
    draw = ImageDraw.Draw(im_parsed)

    for box, shift, thres in [(box1, 0, confidence_threshold1), (box2, w, confidence_threshold2)]:

        for b in box.preorder_traversal():
            if (b.conf is not None) and (b.conf < thres):
                continue
            if isnan(b.text):
                continue

            (x2, y2, w2, h2) = (b.left, b.top, b.width(), b.height())

            fontsize, _ = get_optimal_pillow_font_size(w=w2, h=h2, txt=b.text, d=draw)
            font = ImageFont.truetype(FONT_PATH, fontsize)

            draw.text(xy=(x2 + shift, y2 + h), text=b.text, fill=(0, 0, 0), font=font, spacing=0)

    return np.array(im_parsed)


def _generate_box_type_color_mapping(box: Box) -> Dict[BoxType, Tuple[int, int, int]]:
    used_colors = 0
    mapping: Dict[BoxType, Tuple[int, int, int]] = {}
    for b in box.preorder_traversal():
        if b.box_type not in mapping:
            mapping[b.box_type] = COLORS[used_colors]
            used_colors += 1
    return mapping
